open Ast

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

exception ReturnException of string * string NameMap.t

(* Stores the answers in reverse sequential order *)
let answerList = ref [];;

(* Stores the user choices in resverse sequential order *)
let userChoiceList = ref [];;

(* Removes quotes from the end of a string *)
let rec remove_double_quotes s =
	if( String.get s (String.length s - 1) = '"' &&  String.get s 0 = '"') then 
		String.sub s 1 (String.length s - 2)
	else 
		s;
;;

(* Converts a , delimited string to a list *)
let rec string_to_list x y =
	try
		let i = (String.index y ',' ) in 
			string_to_list ((String.sub y 0 i) :: x) (String.sub y (i+1) ((String.length y) - (i + 1)))
  with Not_found -> y :: x
;;

(* Determines the number of questions that are correct *)
let rec answeredCorrectly correct answers selection =
	if( List.length answers >= 1 && List.hd answers = List.hd selection) then
		answeredCorrectly (correct + 1) (List.tl answers) (List.tl selection)
	else if( List.length answers >= 1 && List.hd answers != List.hd selection) then			
		answeredCorrectly (correct) (List.tl answers) (List.tl selection)
	else
		correct
;;

(* Calculates the percentage of problems correct *)
let rec percentageCorrect answers selection = 
	((float_of_int (answeredCorrectly 0 answers selection)) /. float_of_int (List.length answers)) *. 100.0;;
			
(* Main entry point: run a program *)

let run (vars, funcs) =
  (* Put function declarations in a symbol table *)
  let func_decls = List.fold_left
      (fun funcs fdecl -> NameMap.add fdecl.fname fdecl funcs)
      NameMap.empty funcs
  in
	
  (* Invoke a function and return an updated global symbol table *)
  let rec call fdecl actuals globals =

    (* Evaluate an expression and return (value, updated environment) *)
    let rec eval env = function
				LiteralInteger(i) -> string_of_int(i), env
      | LiteralString(t) -> t, env
			| LiteralBool(b) -> if (b = true) then "1", env else "0", env
			| LiteralFloat(f) -> string_of_float(f), env
			| Noexpr -> string_of_int(1), env (* must be non-zero for the for loop predicate *)
			| Id(var) ->
			  let locals, globals = env in
			  if NameMap.mem var locals then
			    (NameMap.find var locals), env
			  else if NameMap.mem var globals then
			    (NameMap.find var globals), env
			  else raise (Failure ("undeclared identifier " ^ var))
      | Binop(e1, op, e2) ->
				  let v1, env = eval env e1 in
	          let v2, env = eval env e2 in
						  let boolean i = if i then "1" else "0" in
							  (match op with
							    Add ->
										if (String.contains v1 '.') && (String.contains v2 '.') then
											string_of_float((float_of_string v1) +. (float_of_string v2))
										else
											string_of_int((int_of_string v1) + (int_of_string v2))
								| Sub -> 
										if (String.contains v1 '.') && (String.contains v2 '.') then
											string_of_float((float_of_string v1) -. (float_of_string v2))
										else
											string_of_int((int_of_string v1) - (int_of_string v2))
							  | Mult -> 
										if (String.contains v1 '.') && (String.contains v2 '.') then
											string_of_float((float_of_string v1) *. (float_of_string v2))
										else									
											string_of_int((int_of_string v1) * (int_of_string v2))
							  | Div -> 
										if (String.contains v1 '.') && (String.contains v2 '.') then
											string_of_float((float_of_string v1) /. (float_of_string v2))
										else									
											string_of_int((int_of_string v1) / (int_of_string v2))
							  | Equal -> boolean (v1 = v2)
							  | Neq -> boolean (v1 != v2)
							  | Less -> boolean ((int_of_string v1) < (int_of_string v2))
							  | Leq -> boolean ((int_of_string v1) <= (int_of_string v2))
							  | Greater -> boolean ((int_of_string v1) > (int_of_string v2))
							  | Geq -> boolean ((int_of_string v1) >= (int_of_string v2))
								| And -> boolean (((int_of_string v1) = 1) && ((int_of_string v2) = 1))
								| Or -> boolean (((int_of_string v1) = 1) || ((int_of_string v2) = 1))
								| Concat -> (remove_double_quotes v1) ^ (remove_double_quotes v2)
								), env
			| Assign(var, e) ->
			  let v, (locals, globals) = eval env e in
			  if NameMap.mem var locals then
			    v, (NameMap.add var v locals, globals)
			  else if NameMap.mem var globals then
			    v, (locals, NameMap.add var v globals)
			  else raise (Failure ("undeclared identifier " ^ var))
			| Call("print", [e]) ->
			  let v, env = eval env e in print_string (remove_double_quotes v); 
				"0", env
			| Call("println", [e]) ->
			  let v, env = eval env e in print_endline (remove_double_quotes v); 
				"0", env
			| Call("prompt", [e]) ->
				print_string "Question: ";
			  let v, env = eval env e in print_endline (remove_double_quotes v); 
				"0", env
			| Call("answer", [e]) ->
			  let v, env = eval env e in answerList := (remove_double_quotes v) :: !answerList;		
				"0", env	
			| Call("choice", [e]) ->
				print_endline "Choices:";
			  let v, env = eval env e in
					let choices = (List.rev (string_to_list [] (remove_double_quotes v))) in
						List.iter (fun x -> print_endline ("\t" ^ x)) choices;
				print_string "User Choice: ";
				let str = read_line () in 
					userChoiceList := str :: !userChoiceList;
				"0", env
      | Call(f, actuals) ->
			  let fdecl =
			    try NameMap.find f func_decls
			    with Not_found -> raise (Failure ("undefined function " ^ f))
			  in
			  let actuals, env = 
					List.fold_left (fun (actuals, env) actual -> let v, env = eval env actual in v :: actuals, env) ([], env) (List.rev actuals)
			  	in
			  	let (locals, globals) = env in
			  	try
			    	let globals = 
							call fdecl actuals globals in "0", (locals, globals)
						with ReturnException(v, globals) -> v, (locals, globals)
		    	in
	
			    (* Execute a statement and return an updated environment *)
			    let rec exec env = function
							Block(stmts) -> List.fold_left exec env stmts
			      | Expr(e) -> let _, env = eval env e in env
			      | If(e, s1, s2) -> let v, env = eval env e in exec env (if (int_of_string v) != 0 then s1 else s2)
			      | While(e, s) ->
						  let rec loop env =
						    let v, env = eval env e in
						    	if (int_of_string v) != 0 then loop (exec env s) else env
						  			in loop env
			      | For(e1, e2, e3, s) ->
						  let _, env = eval env e1 in
							  let rec loop env =
							    let v, env = eval env e2 in
								    if (int_of_string v) != 0 then
								      let _, env = eval (exec env s) e3 in
								      loop env
								    else
								      env
								  		in loop env 
			      | Return(e) ->
						  let v, (locals, globals) = eval env e in
						  raise (ReturnException(v, globals))
			    in

		    (* Enter the function: bind actual values to formal arguments *)
		    let locals =
		      try 
						List.fold_left2 (fun locals formal actual -> NameMap.add formal actual locals) NameMap.empty fdecl.formals actuals
		      with Invalid_argument(_) -> raise (Failure ("wrong number of arguments passed to " ^ fdecl.fname))
		    in
		    (* Initialize local variables to 0 *)
		    let locals = List.fold_left 
					(fun locals local -> NameMap.add local "0" locals) locals fdecl.locals
		    in
		    (* Execute each statement in sequence, return updated global symbol table *)
		    snd (List.fold_left exec (locals, globals) fdecl.body)
			
			  (* Run a program: initialize global variables to 0, find and run "main" *)
			  in let globals = List.fold_left
			      (fun globals vdecl -> NameMap.add vdecl "0" globals) NameMap.empty vars
			  in try
					let main = call (NameMap.find "main" func_decls) [] globals 
						in 
							if(List.length !answerList > 0) then
								print_endline ("Score:" ^ (string_of_float (percentageCorrect !answerList !userChoiceList) ^ "%"))
			 		with Not_found -> raise (Failure ("did not find the main() function"))